#####################################
#  CoflictSolver
#
#     2017-11-27  by TAKENAKA, Akio

import os.path
import random

from . import conflict_table
from . import local_units
from . import targets


class ConflictSolver():
    """
    Class to solve conflicts among features which are not able to be
    conserved within a site
    """

    def __init__(self):

        self.conflict_table = None  # For an object of ConflictTable
        self.targets = None         # For an object of Target class

        # For an object of LocalUnits within which conflicts are solved.
        self.local_units_adjusted = None  

        # Dictionary to store dispensability index. The keys are feature IDs.
        # The value will be a ditionary, of which keys are values of local units
        #  and the values, corresponding dispensability indices.
        # Initially, values are NONE for all features.
        # 
        self.dispensability_indices = {} 

        # A dictionary to stoe deleted local units to solve conflicts.
        # Keys will be site IDs and values, lists of IDs of features 
        # of deleted local units.

        self.deleted_units = {}  



    def set_local_units(self, obj):
        """
        This function is used to set a LocalUnits object
        
        Parameter
        obj: an object of LocalUnits class

        """

        # For an object of LocalUnits within which conflicts are to be solfved.
        self.local_units_adjusted = obj.dup() 



    def set_conflict_table(self, obj):
        """
        This function is used to set a ConflictTable object
        
        Parameter
        obj: an object of ConflictTable class
        """

        self.conflict_table = obj



    def set_targets(self, obj):
        """
        This function is used to set a Targets object
        
        Parameter
        obj: an object of ConflictTable class
        """

        self.targets = obj



    def solv_conflict(self):
        """
        This function solves conflicts among features found in the original
        LocalUnits object.  New object in which conflicts are solved is returned.
        
        Return
        An object of LocalUnits class.
             It is a newly constructed copy of the original LocalUnits object
             modified to solve all conflicts within.
        """
        
        # Check whether neccessary objects are all set or not.

        if self.local_units_adjusted == None:
            raise(Exception("No LocalUnits object is assigned" 
                          + "before calling ConflictSolver.solv_conflist()"))

        if self.targets == None:
            raise(Exception("No Targets object is assigned" 
                          + "before calling ConflictSolver.solv_conflist()"))

        if self.conflict_table == None:
            raise(Exception("No Required_N object is assigned" 
                          + "before calling ConflictSolver.solv_conflist()"))

        # Solve conflicts

        local_units = self.local_units_adjusted

        sites = local_units.get_sites()

        sites = random.sample(sites, len(sites))

        # Conflicts are solved in sites selected in a random order

        for site in sites:

            features = local_units.get_features_in_site(site)
            conf_features = self.conflict_table.get_features_in_conflict(features)

            if not conf_features: # No features are in conflict with others
                continue

            dispensability_index_dict = {}

            for feature in conf_features:

                v = local_units.get_value_of_feature_in_site(feature, site)
                d_i = self._get_dispensability_index(feature, v)
                dispensability_index_dict[feature] = d_i

            # Features are sorted in the ascending order of dicpensability index
            sorted_features = sorted(conf_features, key = lambda x: dispensability_index_dict[x],
                                               reverse = False)

            # Decide local units to be deleted.
            features_to_del = set()

            while len(sorted_features) >= 2:

                # The first one is the least dispensable among the remaining ones.
                max_val_feature = sorted_features.pop(0)  

                if max_val_feature in features_to_del: # Skip if it is already to be deleted
                    continue

                for feature in sorted_features:  # Check other less indispensable ones.

                    if feature in features_to_del: # Skip if it is already to be deleted
                        continue

                    # If in conflict with the first one, 
                    # add to the list of features to be deleted.

                    if self.conflict_table.in_conflict(max_val_feature, feature) == True:
                        features_to_del.add(feature)

            # Local units to be deleted are removed from the LocalUnits object.

            for feature in features_to_del:
                self._del_local_unit(site, feature)

        # Check completed for all sites

        return self.local_units_adjusted



    def get_deleted_units_in_site(self, site):
        """
        This fuction returns features of local units deleted in the given site.
        
        Parameter
        site: string
            site ID
        
        Return
        list of string:
            IDs of features deleted to solve conflicts in the given site.
        
        """

        if site not in self.deleted_units: # No unit had been deleted
            return list()  # Return an empty list

        else:
            return self.deleted_units[site]


    ######  Following functions are to be used only within this class ########


    def _del_local_unit(self, site, feature):
        """
        This function is used to delete a local unit of given feature 
        in a given site. The copy of the original LocalUnits is modified.

        Parameters
        site: string
            site ID
        feature: string
            feature ID
        """

        self.local_units_adjusted.remove_local_unit(site, feature)

        # Initialize indispensability indices dictionary for this feature.
        # It needs to be newly calculated at the time of the next reference 
        # to the index.
        del self.dispensability_indices[feature]

        # Store local unit deletion info.

        if site not in self.deleted_units:
            self.deleted_units[site] = list()

        self.deleted_units[site].append(feature)



    def _get_dispensability_index(self, feature, value):

        """
        This function returns dispensability index of a local unit
        of a given feature and of a given value.
        
        Parameters
        feature: string
            ID of the feature
        value: numeric
            Value of local units.
            
        Return
        numeric:
            dispensability index of a local unit of a given feature
            of a given value
        """

        if feature not in self.dispensability_indices:
            # Calculate if not found in the dictionary.
            # If found, the found value will be returned.
            v_list = self.local_units_adjusted.get_values_of_feature(feature)

            n = self.targets.get_required_n_for_feature(feature)

            dict = _calc_dispensability_indices(v_list, n)
            self.dispensability_indices[feature] = dict

        if value not in self.dispensability_indices[feature]:
            raise('No index for given feature and value in _get_dispensability_index()')

        return self.dispensability_indices[feature][value]


######  Following functions do not belong to the class ########


def _calc_dispensability_indices(v_list, required_n):
    """
    This function calculates dispensability indices of a list of values.
    Higher value of dispensability index means the corresponding local unit
    is less indispensable to achieve conservation target.
    
    Parameter
    v_list: a list of numeric values
        Values of local units.
    required_n: integer
        number of high valued local units to be covered.

    Return
    dictionary:
        The keys are values of local units, and the values,
        dispensability index of the local units of the individual values.
    """

    dispensability_indices = {} 

    rank_orders = _calc_rank_order(v_list) # Calculate orders for individual values

    for val in rank_orders.keys():  # Calculate indispensability index for each order
        d_index = (rank_orders[val] - 1) / required_n  # following the definition of the index.
        dispensability_indices[val] = d_index

    return dispensability_indices



def _calc_rank_order(v_list):
    """
    This function calculates orders of numeric values in a given list

    Parameter
    v_list: list of numeric values.

    Return
    dictionary:
        The keys are numeric values in v_list and the values, the orders 
        for the values.  If there are ties in the values in v_list, 
        orders for them are not the top among them, but the mean.
        e.g.
        v_list (4, 4, 4, 3, 3, 1)
        dict  (4 => 2, 3 => 4.5, 1 => 6)
    """
    v_list.sort()     # Sort in ascending order
    v_list.reverse()  # descending order

    rank_dict = {}

    cum_n   = 0  # Number of items from the top (including this one)
    cur_val = 0  # Currently focust values
    n_val   = 0  # Numbe of items tying with the current value.

    # Previous valuse.
    # Set a temporal initial value larger than the first one
    prev_v = v_list[0] + 1 

    for v in v_list:

        if v != prev_v:  # If not tying with the prev one,

            if n_val == 0:  # When it is the first of the list
                prev_v = v
                n_val = 1
                cum_n = 1

            else: # Must have tying items
                # Calculate mean order for the previous item and tying ones if any
                mean_rank = (cum_n + (cum_n - n_val + 1))  / 2
                rank_dict[prev_v] = mean_rank

                # prepre for the next item
                prev_v = v
                n_val = 1
                cum_n += 1

        else:  # Tying with the previous one
            n_val += 1
            cum_n += 1

    # Order for the remainings
    mean_rank = (cum_n + (cum_n - n_val + 1))  / 2
    rank_dict[prev_v] = mean_rank

    return rank_dict
