#  ConflictTable
#
#  2017-11-17  by TAKENAKA, Akio

import os.path
import sys
import re


class ConflictTable:
    """
    Class to manage coflicting pairs of features which are not able to be
    conserved within a site
    """

    def __init__(self):

        # Dictionary.  Keys are feature IDs  and the values are lists of feature IDs.
        # If the list for feature1 includes feature2, feature1 and feature2 are
        # in conflict.

        self.table = {}  


    def in_conflict(self, feature1, feature2):
        """
        Parameters
        feature1, feature2: string
            feature IDs. The return is insensitive to the order of the two IDs

        Return
        boolean
            True if feature1 and feature2 are in conflict.
            False if not.
        """
        
        if feature1 not in self.table:
            return False

        if feature2 not in self.table[feature1]:
            return False

        return True


    def register_conflict(self, features_list1, features_list2):
        """
        This function registers conflicting pairs of features.
        
        Parameters
        features_list1, features_list2: lists of feature IDs(strings)
            Any conbination of a feature in list1 and another in list2 is registered
            as a conflicting pair.
            
        """

        # If either of feature list is empty, do nothing

        if not features_list1:  
            return

        if not features_list2:
            return

        # If not list, convert to list
        if not type(features_list1) == list:
            features_list1 = list(features_list1)

        if not type(features_list2) == list:
            features_list2 = list(features_list2)

        # If the key is absent, add the key to the dictionary
        for feature1 in features_list1:
            if feature1 not in self.table:  # If not 
                self.table[feature1] = list()

        for feature2 in features_list2:
            if feature2 not in self.table:
                self.table[feature2] = list()

        for feature1 in features_list1:
            for feature2 in features_list2:
                # Register for both to make in_conflict(f1, f2) 
                # order insensitive.
                if feature2 not in self.table[feature1]:
                    self.table[feature1].append(feature2)
                if feature1 not in self.table[feature2]:
                    self.table[feature2].append(feature1)


    def load_from_file(self, file_name):
        """
        This function loads a conflict data file.
        One or more features in a line are in conflict with the features (one or more)
        in the following line. Feature IDs in a line should be separated by tabs.
        All blank lines are ignored.
        For readability, it is recommended to insert a blank line between pairs of lines
        e.g.
            feature1[tab]feature2
            feature3
        
            feature4[tab]feature5
            feature6[tab]feature7
        
        Parameter
        file_name: string
            Name of the data file
            
        Return
        boolean
            True (if successful) or False (if failed)
        
        """

        if os.path.exists(file_name) == False:
            print("!! ", file_name, " not found in LocalUnits.load_from_file")
            return False

        fin = open(file_name, 'rt', encoding = 'utf-8') 

        # State flag.
        # True when waiting for the 2nd line and False when waiting for the 1st.
        want_2nd_list = False  

        features_list1 = []
        features_list2 = []

        while True:
            line = fin.readline()

            if not line:
                break

            line = line[:-1]

            line = re.sub('#.*$', '', line) #  '#' and what follows areremoved
            line = re.sub('\s+$', '', line) #  Spaces before the end of line is removed

            if re.match('\s*$', line):
                continue             # blank line is skipped

            line = re.sub('\s+', '\t', line) # One or more spaces be replaced by a tabl

            #  A line starts with 0 or more spaces followed by # is skipped
            if re.search('^\s*#.*', line):  
                continue

            if want_2nd_list == True:  # If waiting for the 2nd line,
                features_list2 = line.split('\t')
                self.register_conflict(features_list1, features_list2)
                want_2nd_list = False  # No more waiting for the 2nd line

            else:  # If not waiting for the 2nd (i.e., wating for the 1st),
                features_list1 = line.split('\t')
                want_2nd_list = True  # Ready for the 2nd line

        fin.close()

        return True


    def get_features_in_conflict(self, features):
        """
        This function finds features which are in conflict with others in the list
        
        Parameter
        features: list of feature IDs (string)
            Conflicting pairs within this list is searched.
            
        Return
        list of string
            IDs of features in the list, which is in conflict with one or more 
            features in the list.
        """

        # All features registered
        registered_features = self.table.keys() 

        # A set to store conflicting features.
        # It is a set, not list, to avoid replicated appending of the same ID.
        conflicting_features = set()  

        n_features = len(features)

        for i in range(0, n_features - 1):

            if features[i] not in registered_features:
                # features[i] is in conflict with none of the given features.
                continue

            for j in range(i + 1, n_features):
                if self.in_conflict(features[i], features[j]):
                    # register both feature[i] and feature[j]
                    conflicting_features.add(features[i])
                    conflicting_features.add(features[j])

        return list(conflicting_features) # Convert a set to a list.
