#  ScatterPenalty
#
#     2020-03-18  by TAKENAKA, Akio

import os.path
import sys
import re

# Import implementation class which calculates scatter penalty
from . import boundary_penalty
from . import distance_to_nearest


class ScatterPenalty:
    """
    This class is a wrapper of an implementation class calculating penalty 
    of scattered sites. The implementation class should have the following 
    three methods.

    get_penalty(self, sites, site_to_add)
        Return
            float: penalty by adding 'site_to_add' to 'sites'.

    load_data(self, file_name)
        Return
            boolean:  True when the file is loaded successfully.
                   The file contains data for calculation of penalty.

    is_ready(self)
        Return
            boolean: True when the object is ready to evaluate penaly.

    """

    def __init__(self):

        self.penalty_obj = None  # 

        return


    def choose_algorithm(self, algorithm):

        """
        This function sets scatter penalty algorithm to be used.
        If the given algorithm is not available, False is returned.
        The implementation class of the algorithm should be imported in this file.

        Parameter
        algorithm: string
            A string specifying algorithm to be used. case insensitive.

        Return
        boolean
            True if successful, False if failed due to invalid algorithm name
        """

        if re.match("BOUNDARY_LEN", algorithm, re.IGNORECASE):
            self.penalty_obj = boundary_penalty.BoundaryPenalty()
            # print("# BOUNDARY_LEN")
            return True

        elif re.match("DISTANCE_TO_NEAREST", algorithm, re.IGNORECASE):
            self.penalty_obj = distance_to_nearest.DistanceToNearest()
            # print("DISTANCE_TO_NEAREST")
            return True

            '''
            elif re.match("XXXXX", algorithm, re.IGNORECASE):
                self.penalty_obj = xxxxxxx.XXXXXXX()
                return True
            '''

        else:  # given algorithm does not match available ones
            return False


    def get_penalty(self, sites, site_to_add):
        """
        This function returns the scatter penalty when a site is 
        added to a set of sites.  The parameters are transfered to the 
        implementation class object

        Parameters
        sites: list of strings
            a list of IDs of already selected sites
        site_to_add: string
            The ID of the new site to be added

        Return
        float
            Penalty when 'site_to_add' is added to 'sites'.
        """

        if self.penalty_obj.is_ready() == False:
            print("!! scatter_penalty object is not ready")
            sys.exit()

        penalty = self.penalty_obj.get_penalty(sites, site_to_add)

        return penalty


    def load_data(self, file_name):

        """
        This function loads data needed for penalty calculation from a file 
        The contents of the file depends on the implementation class used.

        Parameter
        file_name: string
            Name of a file to be loaded

        Return
        boolean
            True if loaded successfully and False, if failed.
        """

        if os.path.exists(file_name) is not True:
            print("!! ", file_name, " not found in ScatterPenalty.load_penalty_data")
            return False

        rvl = self.penalty_obj.load_data(file_name)

        return rvl

