#  SecSel_Core
#
#  2017-11-17  by TAKENAKA, Akio


from . import local_units
from . import cosspl
from . import scatter_penalty



class SecSel_Core():
    """
    Core class of site selection.
    """

    def __init__(self):

        self.local_units = None    # Place holder of a LocalUnits object

        self.top_n_upto       = 1   # Try to cover this Top-N

        self.reserved_sites = list()  # Pre-reserved sites

        self.present_PA = list()    # Present protected area (list of sites)

        # Allowed upset of less effective sites when they are superior in 
        # avoiding costs and scatter penalties.  If set 0, no upset is allowed.
        self.max_upset = 0

        # SiteCost class oject.
        self.site_cost = None

        # Object of a class to calculate scattered sites penalty
        self.scatter_penalty_obj = None 

        self.consider_cost_first = False  # consider cost before scatter penaly.
                                          # By default, False (penalty first)



    def set_top_n_upto(self, top_n):
        """
        Function to set conservation taret, 
        number of high-value local units are to be covered.
        
        Parameter
        top_n: integer
            number of high-value local units are to be covered.
        """

        self.top_n_upto       = top_n



    def set_reserved_sites(self, sites):
        """
        Function to set pre-reserved sites. 
        The site selection starts with these sites.
        
        Parameter
        sites: list of strings
            list of IDs of sites to be pre-reserved.
        """
        
        self.reserved_sites = sites



    def set_max_upset(self, upset):
        """
        Function to set allowed upset of less effective sites when they are superior
        in avoiding costs and scatter penalties.  If set 0, no upset is allowed.

        Parameter
        upset: integer
            allowed upset
        """

        self.max_upset = upset



    def set_scatter_penalty_obj(self, obj):
        """
        Function to set an object to calculate scatterd site penalty.
        
        Parameter
        obj: object of ScatterPenalty class
        """
        self.scatter_penalty_obj = obj



    def set_to_consider_cost_first(self, cost_first = True):
        """
        Functionto set priority on site cost over the scatter penalty
        to break ties in the site selection process.
        
        Parameter
        const_first: boolean
            True, to set pririty on cost over the scatter penalty.
            This parameter can be omitted. 
        """
        self.consider_cost_first = cost_first



    def set_site_cost(self, obj):
        """
        Function to set an object of SiteCost
        
        Parameter
        obj: object of SiteCost class
        """
        
        self.site_cost = obj



    def set_local_units(self, obj):
        """
        Function to set an object of LocalUnits
        
        Parameter
        obj: object of LocalUnits class
        """

        self.local_units = obj



    def get_selected_sites(self):
        """
        Function to get selected sites
        
        Return
        list of strings:
            list of IDs of selected sites in the selected order.
        """

        return list(self.present_PA)



    def select_sites(self):
        """
        Function to select sites. This function should be called after all 
        necessary parameters and objects are set.
        
        Return
        list of strings:
            list of IDs of selected sites in the selected order.
            The same list can be obtained afterwards by calling
            get_selected_sites()
        """

        # Outline of the algorithm of this function is as follows:
        #
        # Consider an array, V, which consists of values of local units of 
        # a feature sorted in descending order (e.g., k, k, k-1, k-1, k-2, 
        # ..., 2, 1, 1).
        # If a set of sites include local units of a feature, of which values 
        # include top n of the above array V, the feature is defined to be 
        # top-n included in the set of sites.
        #
        # Assume that the maximum value in all V's of individual features is K.  
        # First, this function selects sites to cover features which has local 
        # units of the value K. The selected sites should include one or more 
        # local units of value K for all the features. The site selection 
        # process is left to CoSSPL object, in which the greedy algorithm 
        # is applied.  The set of selected sites will satisfy 'top-1 included' 
        # condition for all the features with local units of value K.
        # 
        # Next, consider features of which the maxim value of local units is K-1.
        # This function add sites to include one or more local units of value k-1 
        # for all the features.  As a result, we obtain site sets which satisfy 
        # 'top-1 included' condition for all features of which the maximum value 
        #  of local units is k-1 or higher.
        #
        # Repeating the above decreasing value down to 1, we obtain a set of sites 
        # which satisfies "top-1 included" condition for all features.
        # 
        # Then, repeat all of the above process for the second element of V 
        # of each feature to add more sites to the set. We will obtain a set of
        # sites which satisfy 'top-2 included' condition for all features with
        # two or more local units.
        
        # We continue the above to the third, fourth, 5th, ... n7 th element of 
        # V's. In the end, a set of sites which satisfy J 'top-n condition' for
        # all features with n or more local units.


        self.present_PA = list()    # List of currently selected sites

        max_value = self.local_units.get_highest_value()

        features = self.local_units.get_features() #  List of all feature IDs

        # Dictionary.  Keys will be feature IDs, and values will be lists of values
        # of local units of individual features sorted in descending order.
        sorted_values = {}

        for feature in features:

            # Values of local units of thefeatreu
            values = self.local_units.get_values_of_feature(feature)

            values.sort()    # values are in ascending order
            values.reverse() # values are in descending order

            sorted_values[feature] = values; # Stored in the dictionary

        # Defined before the following loop starts, because the value will be accessed after the loop
        i_top = 1 

        for i_top in range(1, self.top_n_upto + 1):   # n-top loop

            for target_value in range(max_value, 0, -1):   #  Starts from the max value

                coss = cosspl.CoSSPL()    # CoSSPL object 

                coss.set_max_upset(self.max_upset)
                coss.set_site_cost(self.site_cost)

                if self.consider_cost_first == True:
                    coss.set_to_consider_cost_first(True)
                else:
                    coss.set_to_consider_cost_first(False)

                coss.set_scatter_penalty_obj(self.scatter_penalty_obj)

                # Set pre-reserved sites (if any)
                if len(self.reserved_sites) > 0:
                    coss.set_reserved_sites(self.reserved_sites)

                # Then, add so far selected sites as pre-reserved sites.
                if len(self.present_PA) > 0:  
                    coss.add_reserved_sites(self.present_PA)

                #########

                n_units_added = 0  # Number of registered local units

                for feature in features:  # Loop for features

                    # Values of local units of the feature, sorted in descending order.
                    values = sorted_values[feature] 

                    # If the number of local units of the feature is less than i_top,
                    # the feature is skipped.

                    if len(values) < i_top:  
                        continue

                    if values[i_top - 1]  != target_value:  
                        # Skip if i_top'th value (counted from 1) is not the target value
                        continue

                    for value_work in range(max_value, target_value - 1, -1):

                        sites = self.local_units.get_sites_with_feature_of_given_value(
                                                                    feature, value_work)

                        if len(sites) == 0:
                            continue

                        for site in sites:
                            coss.add_feature_site_record(feature, site)
                            n_units_added += 1

                if n_units_added == 0:  
                    # No local unit of this value is added, so no job forCoSSPL object
                    continue

                # set target, i.e., top i_top included
                coss.set_required_occurrence(i_top)  

                coss.select_sites() # Execution of site selection (by a CoSSPL object)

                # Tuple of IDs of selected sites in the order of selection
                selected_sites = coss.get_selected_sites()

                self.present_PA = list(selected_sites);

            ### End of target_value loop

        ### End of n_top loop

        return self.present_PA
